/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SeqScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;              // 表的名称
    std::vector<Condition> conds_;      // scan的条件
    RmFileHandle *fh_;                  // 表的数据文件句柄
    std::vector<ColMeta> cols_;         // scan后生成的记录的字段
    size_t len_;                        // scan后生成的每条记录的长度
    std::vector<Condition> fed_conds_;  // 同conds_，两个字段相同

    Rid rid_;
    std::unique_ptr<RecScan> scan_;     // table_iterator

    SmManager *sm_manager_;

   public:
    SeqScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = std::move(tab_name);
        conds_ = std::move(conds);
        TabMeta &tab = sm_manager_->db_.get_table(tab_name_);
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab.cols;
        len_ = cols_.back().offset + cols_.back().len;

        context_ = context;

        fed_conds_ = conds_;
    }
    size_t tupleLen() const override { return len_; }
    void beginTuple() override {
        scan_ = std::make_unique<RmScan>(fh_);
        while(!scan_->is_end()){
            rid_ = scan_->rid();
            try{
                auto record = fh_->get_record(rid_, context_);
                if(eval_conds(cols_, conds_, record.get())){
                    //若找到满足条件的元组，则终止
                    break;
                }
            }catch(RecordNotFoundError &e){
                std::cerr << e.what() << std::endl;
            }
            scan_->next();
        }
    }
    
    bool is_end() const override { return scan_->is_end(); }

    void nextTuple() override {
        scan_->next();
        while(!scan_->is_end()){
            rid_ = scan_->rid();
            try{
                auto record = fh_->get_record(rid_, context_);
                if(eval_conds(cols_, conds_, record.get())){
                    break;
                }
            }catch(RecordNotFoundError &e){
                std::cerr << e.what() << std::endl;
            }
            scan_->next();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        return fh_->get_record(rid_, context_);
    }

    Rid &rid() override { return rid_; }
    bool eval_cond(const std::vector<ColMeta> &rec_cols, const Condition &cond, const RmRecord *rec) {
        // 获取左操作数
        auto lhs_col = get_col(rec_cols, cond.lhs_col);
        char *lhs = rec->data + lhs_col->offset;

        // 获取右操作数
        char *rhs;
        ColType rhs_type;
        if (cond.is_rhs_val) {
            //右操作数是值
            rhs_type = cond.rhs_val.type;
            rhs = cond.rhs_val.raw->data;
        } else {
            // 右操作数是列
            auto rhs_col = get_col(rec_cols, cond.rhs_col);
            rhs_type = rhs_col->type;
            rhs = rec->data + rhs_col->offset;
        }

        // 断言左右操作数的类型相同
        assert(rhs_type == lhs_col->type);

        // 比较操作
        int cmp = ix_compare(lhs, rhs, rhs_type, lhs_col->len);

        // 根据操作符进行条件判断
        switch (cond.op) {
            case OP_EQ:
                return cmp == 0;
            case OP_NE:
                return cmp != 0;
            case OP_LT:
                return cmp < 0;
            case OP_GT:
                return cmp > 0;
            case OP_LE:
                return cmp <= 0;
            case OP_GE:
                return cmp >= 0;
            default:
                throw InternalError("Unexpected op type");
        }
    }
    const std::vector<ColMeta> &cols() const override { return cols_; }

    //这段代码的作用是在给定的记录 rec 中，评估一组条件 conds 是否全部成立。它利用了 C++ STL 中的 std::all_of 算法来遍历并评估每个条件，并返回整体的评估结果。
    bool eval_conds(const std::vector<ColMeta> &rec_cols, const std::vector<Condition> &conds, const RmRecord *rec) {
        return std::all_of(conds.begin(), conds.end(),
                           [&](const Condition &cond) { return eval_cond(rec_cols, cond, rec); });
    }
};